import logging

import google.api_core.exceptions
from google.cloud.pubsub_v1 import SubscriberClient
from google.protobuf.duration_pb2 import Duration  # noqa
from google.protobuf.field_mask_pb2 import FieldMask  # noqa
from google.pubsub_v1.types.pubsub import (
    ExpirationPolicy,
    PushConfig,
    RetryPolicy,
    Subscription as _Subscription,
    UpdateSubscriptionRequest,
)

from octue.cloud.service_id import OCTUE_SERVICES_NAMESPACE


logger = logging.getLogger(__name__)

# Useful time periods in seconds.
THIRTY_ONE_DAYS = 31 * 24 * 3600


class Subscription:
    """A candidate subscription to use with Google Pub/Sub. The subscription represented by an instance of this class
    does not necessarily already exist on the Google Pub/Sub servers.

    :param str name: the name of the subscription excluding "projects/<project_name>/subscriptions/<namespace>"
    :param octue.cloud.pub_sub.topic.Topic topic: the topic the subscription is attached to
    :param str project_name: the name of the Google Cloud project that the subscription belongs to
    :param int ack_deadline: the time in seconds after which, if the subscriber hasn't acknowledged a message, to retry sending it to the subscription
    :param int message_retention_duration: unacknowledged message retention time in seconds
    :param int|float|None expiration_time: number of seconds of inactivity after which the subscription is deleted (infinite time if `None`)
    :param float minimum_retry_backoff: minimum number of seconds after the acknowledgement deadline has passed to exponentially retry delivering a message to the subscription
    :param float maximum_retry_backoff: maximum number of seconds after the acknowledgement deadline has passed to exponentially retry delivering a message to the subscription
    :param str|None push_endpoint: if this is a push subscription, this is the URL to which messages should be pushed; leave as `None` if this is a pull subscription
    :return None:
    """

    def __init__(
        self,
        name,
        topic,
        project_name,
        ack_deadline=600,
        message_retention_duration=600,
        expiration_time=THIRTY_ONE_DAYS,
        minimum_retry_backoff=10,
        maximum_retry_backoff=600,
        push_endpoint=None,
    ):
        if not name.startswith(OCTUE_SERVICES_NAMESPACE):
            self.name = f"{OCTUE_SERVICES_NAMESPACE}.{name}"
        else:
            self.name = name

        self.topic = topic
        self.path = self.generate_subscription_path(project_name, self.name)
        self.ack_deadline = ack_deadline
        self.message_retention_duration = Duration(seconds=message_retention_duration)

        # If expiration_time is `None`, the subscription will never expire.
        if expiration_time is None:
            self.expiration_policy = ExpirationPolicy(mapping=None)
        else:
            self.expiration_policy = ExpirationPolicy(mapping=None, ttl=Duration(seconds=expiration_time))

        self.retry_policy = RetryPolicy(
            mapping=None,
            minimum_backoff=Duration(seconds=minimum_retry_backoff),
            maximum_backoff=Duration(seconds=maximum_retry_backoff),
        )

        self.push_endpoint = push_endpoint
        self._subscriber = SubscriberClient()
        self._created = False

    @property
    def creation_triggered_locally(self):
        """Was the subscription successfully created by calling `self.create` locally? This is `False` if its creation
        was triggered remotely.

        :return bool:
        """
        return self._created

    @property
    def is_pull_subscription(self):
        """Return `True` if this is a pull subscription.

        :return bool:
        """
        return self.push_endpoint is None

    @property
    def is_push_subscription(self):
        """Return `True` if this is a push subscription.

        :return bool:
        """
        return self.push_endpoint is not None

    def __repr__(self):
        """Represent the subscription as a string.

        :return str:
        """
        return f"<{type(self).__name__}({self.name})>"

    def create(self, allow_existing=False):
        """Create a Google Pub/Sub subscription that can be subscribed to.

        :param bool allow_existing: if `False`, raise an error if the subscription already exists; if `True`, do nothing (the existing subscription is not overwritten)
        :return google.pubsub_v1.types.pubsub.Subscription:
        """
        subscription = self._create_proto_message_subscription()

        if not allow_existing:
            subscription = self._subscriber.create_subscription(request=subscription)
            self._created = True
            self._log_creation()
            return subscription

        try:
            subscription = self._subscriber.create_subscription(request=subscription)
            self._created = True
        except google.api_core.exceptions.AlreadyExists:
            pass

        self._log_creation()
        return subscription

    def update(self):
        """Update an existing subscription with the state of this instance.

        :return None:
        """
        self._subscriber.update_subscription(
            request=UpdateSubscriptionRequest(
                mapping=None,
                subscription=self._create_proto_message_subscription(),  # noqa
                update_mask=FieldMask(
                    paths=["ack_deadline_seconds", "message_retention_duration", "expiration_policy", "retry_policy"]
                ),
            )
        )

    def delete(self):
        """Delete the subscription from Google Pub/Sub.

        :return None:
        """
        self._subscriber.delete_subscription(subscription=self.path)
        logger.info("Subscription %r deleted.", self.path)

    def exists(self, timeout=5):
        """Check if the subscription exists on the Google Pub/Sub servers.

        :param float timeout:
        :return bool:
        """
        try:
            self._subscriber.get_subscription(subscription=self.path, timeout=timeout)
            return True
        except google.api_core.exceptions.NotFound:
            return False

    @staticmethod
    def generate_subscription_path(project_name, subscription_name):
        """Generate a full subscription path in the format `projects/<project_name>/subscriptions/<subscription_name>`.

        :param str project_name:
        :param str subscription_name:
        :return str:
        """
        return f"projects/{project_name}/subscriptions/{subscription_name}"

    def _create_proto_message_subscription(self):
        """Create a Proto message subscription from the instance to be sent to the Pub/Sub API.

        :return google.pubsub_v1.types.pubsub.Subscription:
        """
        if self.push_endpoint:
            push_config = {"push_config": PushConfig(mapping=None, push_endpoint=self.push_endpoint)}  # noqa
        else:
            push_config = {}

        return _Subscription(
            mapping=None,
            name=self.path,  # noqa
            topic=self.topic.path,
            ack_deadline_seconds=self.ack_deadline,  # noqa
            message_retention_duration=self.message_retention_duration,  # noqa
            expiration_policy=self.expiration_policy,  # noqa
            retry_policy=self.retry_policy,  # noqa
            **push_config,
        )

    def _log_creation(self):
        """Log the creation of the subscription.

        :return None:
        """
        logger.debug("Subscription %r created.", self.path)



import pandas


def read_csv_files(files):
    """Read a sequence of CSV files file containing meteorological mast anemometer and wind vane data

    You don't really need to care about this, because your files are unlikely to be in the same form as our
    example csv files. But for the sake of a complete example, we show you how we'd read these in here.

    :param list(octue.resources.Datafile) files: List of the file names to read in and concatenate
    :return pandas.DataFrame: Pandas dataframe containing the imported, uncleaned data
    """
    # This is a simple concatenation. If you have a huge dataset, it's worth getting into working with remote files on
    # the cloud and/or doing this in batches.
    frames = []
    for file in files:
        frames.append(pandas.read_csv(file.local_path))

    return pandas.concat(frames)



VALUES_FILENAME = "values.json"
MANIFEST_FILENAME = "manifest.json"

STRAND_FILENAME_MAP = {
    "configuration_values": VALUES_FILENAME,
    "configuration_manifest": MANIFEST_FILENAME,
    "input_values": VALUES_FILENAME,
    "input_manifest": MANIFEST_FILENAME,
    "output_values": VALUES_FILENAME,
    "output_manifest": MANIFEST_FILENAME,
}

# TODO this should probably be defined in twined
OUTPUT_STRANDS = ("output_values", "output_manifest")

# TODO this should probably be defined in twined
RUN_STRANDS = ("input_values", "input_manifest", "credentials", "children")

GOOGLE_COMPUTE_PROVIDERS = {"GOOGLE_CLOUD_FUNCTION", "GOOGLE_CLOUD_RUN", "GOOGLE_DATAFLOW"}



from abc import ABC
from collections import UserDict

from octue import exceptions
from octue.mixins import Filterable
from octue.utils.objects import get_nested_attribute


class FilterContainer(ABC):
    """A mixin for containers to allow their elements to be filtered and/or ordered according to the elements'
    attributes. Mix with a primitive python container type (e.g. list, set, dict) to get a filterable version of it.

    Key points:
    - Any attribute of a member of a filter container whose type or interface is supported can be used when filtering
    - Filters are named as "<name_of_attribute_to_check>__<filter_action>"
    - Multiple filters can be specified at once for chained filtering
    - ``<name_of_attribute_to_check>`` can be a single attribute name or a double-underscore-separated string of nested
      attribute names
    - Nested attribute names work for real attributes as well as dictionary keys (in any combination and to any depth)

    For a ``FilterContainer`` instance to work, all its members must inherit from ``octue.mixins.filterable.Filterable``.

    .. code-block:: python

        from octue.resources import Datafile

        class FilterSet(FilterContainer, set):
            pass

        filter_set = FilterSet(
            {
                Datafile(path="my_file.csv", tags={"cluster": 0, "manufacturer": "Vestas"}),
                Datafile(path="your_file.txt", tags={"cluster": 1, "manufacturer": "Vergnet"}),
                Datafile(path="another_file.csv", tags={"cluster": 2, "manufacturer": "Enercon"})
            }
        )

        # Single filter, non-nested attribute.
        filter_set.filter(name__ends_with=".csv")
        >>> <FilterSet({<Datafile('my_file.csv')>, <Datafile('another_file.csv')>})>

        # Two filters, non-nested attributes.
        filter_set.filter(name__ends_with=".csv", tags__cluster__gt=1)
        >>> <FilterSet({<Datafile('another_file.csv')>})>

        # Single filter, nested attribute.
        filter_set.filter(tags__manufacturer__startswith("V"))
        >>> <FilterSet({<Datafile('my_file.csv')>, <Datafile('your_file.csv')>})>
    """

    def filter(self, ignore_items_without_attribute=True, **kwargs):
        """Return a new instance containing only the `Filterable`s to which the given filter criteria are `True`.

        :param bool ignore_items_without_attribute: if True, just ignore any members of the container without a filtered-for attribute rather than raising an error
        :param {str: any} kwargs: keyword arguments whose keys are the name of the filter and whose values are the values to filter for
        :return octue.resources.filter_containers.FilterContainer:
        """
        if any(not isinstance(item, Filterable) for item in self):
            raise TypeError(f"All items in a {type(self).__name__} must be of type {Filterable.__name__}.")

        raise_error_if_filter_is_invalid = not ignore_items_without_attribute

        if len(kwargs) == 1:
            return type(self)(
                (
                    item
                    for item in self
                    if item.satisfies(raise_error_if_filter_is_invalid=raise_error_if_filter_is_invalid, **kwargs)
                )
            )

        filter_names = list(kwargs)

        for filter_name in filter_names:
            filter_value = kwargs.pop(filter_name)
            return self.filter(raise_error_if_filter_is_invalid, **{filter_name: filter_value}).filter(**kwargs)

    def order_by(self, attribute_name, check_start_value=None, check_constant_increment=None, reverse=False):
        """Order the `Filterable`s in the container by an attribute with the given name, returning them as a new
        `FilterList` regardless of the type of filter container begun with (`FilterSet`s and `FilterDict`s are
        inherently orderless).

        :param str attribute_name: name of attribute (optionally nested) to order by e.g. "a", "a.b", "a.b.c"
        :param any check_start_value: if provided, check that the first item in the ordered container has the given start value for the attribute ordered by
        :param int|float|None check_constant_increment: if given, check that the ordered-by attribute of each of the items in the ordered container increases by the given value when progressing along the sequence
        :param bool reverse: if True, reverse the ordering
        :raise octue.exceptions.InvalidInputException: if an attribute with the given name doesn't exist on any of the container's members
        :return FilterList:
        """
        attribute_name = ".".join(attribute_name.split("__"))

        try:
            results = FilterList(
                sorted(self, key=lambda item: get_nested_attribute(item, attribute_name), reverse=reverse)
            )

        except AttributeError:
            raise exceptions.InvalidInputException(
                f"An attribute named {attribute_name!r} does not exist on one or more members of {self!r}."
            )

        if check_start_value is not None:
            if get_nested_attribute(results[0], attribute_name) != check_start_value:
                raise exceptions.BrokenSequenceException(
                    f"The attribute {attribute_name!r} of the first item of {results!r} does equal the given start "
                    f"value {check_start_value!r}."
                )

        if check_constant_increment is not None:
            required_increment = check_constant_increment

            for i in range(len(results) - 1):
                actual_increment = get_nested_attribute(results[i + 1], attribute_name) - get_nested_attribute(
                    results[i], attribute_name
                )

                if actual_increment != required_increment:
                    raise exceptions.BrokenSequenceException(
                        f"The attributes {attribute_name!r} of the items of {results!r} do not increase by a constant "
                        f"increment of {required_increment}."
                    )

        return results

    def one(self, **kwargs):
        """If a single result exists for the given filters, return it. Otherwise, raise an error.

        :param {str: any} kwargs: keyword arguments whose keys are the name of the filter and whose values are the values to filter for
        :raise octue.exceptions.UnexpectedNumberOfResultsException: if zero or more than one results satisfy the filters
        :return octue.resources.mixins.filterable.Filterable:
        """
        results = self.filter(**kwargs)

        # If no filters are given, results will be `None`.
        if results is None:
            self._raise_if_not_exactly_one_item(self, **kwargs)
            return next(iter(self))

        self._raise_if_not_exactly_one_item(results, **kwargs)
        return results.pop()

    def _raise_if_not_exactly_one_item(self, iterable, **kwargs):
        """Raise an error if the given iterable doesn't have exactly one item.

        :param iter iterable:
        :param kwargs: key-value pairs of filters used to produce the iterable (to add information to the error message)
        :raise octue.exceptions.UnexpectedNumberOfResultsException: if the iterable doesn't have exactly one item
        :return None:
        """
        if len(iterable) > 1:
            raise exceptions.UnexpectedNumberOfResultsException(f"More than one result found for filters {kwargs}.")

        if len(iterable) == 0:
            raise exceptions.UnexpectedNumberOfResultsException(f"No results found for filters {kwargs}.")


class FilterSet(FilterContainer, set):
    pass


class FilterList(FilterContainer, list):
    pass


class FilterDict(FilterContainer, UserDict):
    """A dictionary that is filterable by its values' attributes. Each key can be anything, but each value must be an
    ``octue.mixins.filterable.Filterable`` instance.
    """

    def filter(self, ignore_items_without_attribute=True, **kwargs):
        """Return a new instance containing only the Filterables for which the given filter criteria apply are
        satisfied.

        :param bool ignore_items_without_attribute: if True, just ignore any members of the container without a filtered-for attribute rather than raising an error
        :param {str: any} kwargs: keyword arguments whose keys are the name of the filter and whose values are the values to filter for
        :return FilterDict:
        """
        if any(not isinstance(item, Filterable) for item in self.values()):
            raise TypeError(f"All values in a {type(self).__name__} must be of type {Filterable.__name__}.")

        raise_error_if_filter_is_invalid = not ignore_items_without_attribute

        if len(kwargs) == 1:
            return type(self)(
                {
                    key: value
                    for key, value in self.items()
                    if value.satisfies(raise_error_if_filter_is_invalid=raise_error_if_filter_is_invalid, **kwargs)
                }
            )

        filter_names = list(kwargs)

        for filter_name in filter_names:
            filter_value = kwargs.pop(filter_name)
            return self.filter(raise_error_if_filter_is_invalid, **{filter_name: filter_value}).filter(**kwargs)

    def order_by(self, attribute_name, reverse=False):
        """Order the instance by the given attribute_name, returning the instance's elements as a new FilterList.

        :param str attribute_name: name of attribute (optionally nested) to order by e.g. "a", "a.b", "a.b.c"
        :param bool reverse: if True, reverse the ordering
        :raise octue.exceptions.InvalidInputException: if an attribute with the given name doesn't exist on any of the FilterDict's values
        :return FilterList:
        """
        try:
            return FilterList(
                sorted(self.items(), key=lambda item: get_nested_attribute(item[1], attribute_name), reverse=reverse)
            )

        except AttributeError:
            raise exceptions.InvalidInputException(
                f"An attribute named {attribute_name!r} does not exist on one or more members of {self!r}."
            )

    def one(self, **kwargs):
        """If a single item exists for the given filters, return it. Otherwise, raise an error.

        :param {str: any} kwargs: keyword arguments whose keys are the name of the filter and whose values are the values to filter for
        :raise octue.exceptions.UnexpectedNumberOfResultsException: if zero or more than one results satisfy the filters
        :return (any, octue.resources.mixins.filterable.Filterable):
        """
        results = self.filter(**kwargs)

        # If no filters are given, results will be `None`.
        if results is None:
            self._raise_if_not_exactly_one_item(self, **kwargs)
            return next(iter(self.items()))

        self._raise_if_not_exactly_one_item(results, **kwargs)
        return results.popitem()



"""The names of backend classes must end with "Backend" to be registered as valid backends for the `Child` class.
Beyond that, a backend class can store whatever data it needs to authenticate and specify resources for the specific
backend it represents. Credentials should not be stored directly in a backend instance, but storing the names of
environment variables that store credentials is fine. For a backend to be valid, it must have a corresponding entry in
the "oneOf" field of the "backend" key of the children schema in `Twined`, which is located at
`twined/schema/children_schema.json`.
"""
from abc import ABC

from octue import exceptions


def get_backend(backend_name=None):
    """Get the service backend with the given name.

    :param str|None backend_name: if `None`, return the `GCPPubSubBackend`
    :return ServiceBackend:
    """
    if not backend_name:
        return GCPPubSubBackend

    if backend_name not in AVAILABLE_BACKENDS:
        raise exceptions.BackendNotFound(
            f"Backend with name {backend_name} not found. Available backends are {list(AVAILABLE_BACKENDS.keys())}"
        )

    return AVAILABLE_BACKENDS[backend_name]


class ServiceBackend(ABC):
    """A dataclass specifying the backend for an Octue Service, including any credentials and other information it
    needs.

    :return None:
    """


class GCPPubSubBackend(ServiceBackend):
    """A dataclass containing the details needed to use Google Cloud Platform Pub/Sub as a Service backend.

    :param str project_name:
    :return None:
    """

    def __init__(self, project_name):
        if project_name is None:
            raise exceptions.CloudLocationNotSpecified(
                "The project name must be specified for a service to connect to the correct Google Cloud Pub/Sub "
                f"instance - it's currently {project_name!r}.",
            )

        self.project_name = project_name

    def __repr__(self):
        return f"<{type(self).__name__}(project_name={self.project_name!r})>"


AVAILABLE_BACKENDS = {
    key: value for key, value in locals().items() if key.endswith("Backend") and key != "ServiceBackend"
}



# The functions in this module were based on https://github.com/brutasse/rgb2ansi.

import binascii


WHITE = "ffffff"


def colourise(string, text_colour=WHITE, background_colour=None):
    """Colour in a string with the given text colour and, optionally, background colour.

    :param str string: the string to colourise
    :param str text_colour: the hex representation of the text colour
    :param str|None background_colour: the hex representation of the background colour
    :return str: the colourised string
    """
    text_colour = _convert_colour_to_ansi(text_colour)

    if not background_colour:
        return f"\x1b[38;5;{text_colour}m{string}\x1b[0m"

    background_colour = _convert_colour_to_ansi(background_colour)
    return f"\x1b[38;5;{text_colour};48;5;{background_colour}m{string}\x1b[0m"


def _convert_colour_to_ansi(colour):
    if isinstance(colour, str):
        colour = binascii.a2b_hex(colour)

    if isinstance(colour, bytes):
        colour = list(map(int, colour))

    red, green, blue = list(map(_round_colour, colour))
    return 16 + 36 * red + 6 * green + blue


def _round_colour(colour):
    quot = 255 / 5
    return round(colour / quot)



import concurrent.futures
import copy

from octue.cloud.pub_sub.service import Service
from octue.resources import service_backends


BACKEND_TO_SERVICE_MAPPING = {"GCPPubSubBackend": Service}


class Child:
    """A class representing an Octue child service that can be asked questions. This is a convenience wrapper for
    `Service` that makes asking questions more intuitive and allows easier selection of backends.

    :param str id: the ID of the child
    :param dict backend: must include the key "name" with a value of the name of the type of backend e.g. "GCPPubSubBackend" and key-value pairs for any other parameters the chosen backend expects
    :param str internal_service_name: the name to give to the internal service used to ask questions to the child
    :return None:
    """

    def __init__(self, id, backend, internal_service_name="local/local:local"):
        self.id = id

        backend = copy.deepcopy(backend)
        backend_type_name = backend.pop("name")
        backend = service_backends.get_backend(backend_type_name)(**backend)

        self._service = BACKEND_TO_SERVICE_MAPPING[backend_type_name](name=internal_service_name, backend=backend)

    def __repr__(self):
        """Represent the child as a string.

        :return str:
        """
        return f"<{type(self).__name__}({self.id!r})>"

    @property
    def received_messages(self):
        """Get the messages received from the child if it has been asked a question. If it hasn't, `None` is returned.
        If an empty list is returned, no messages have been received.

        :return list(dict)|None:
        """
        return self._service.received_messages

    def ask(
        self,
        input_values=None,
        input_manifest=None,
        children=None,
        subscribe_to_logs=True,
        allow_local_files=False,
        handle_monitor_message=None,
        record_messages=True,
        allow_save_diagnostics_data_on_crash=True,
        question_uuid=None,
        timeout=86400,
        maximum_heartbeat_interval=300,
    ):
        """Ask the child a question and wait for its answer - i.e. send it input values and/or an input manifest and
        wait for it to analyse them and return output values and/or an output manifest. The input values and manifest
        must conform to the schema in the child's twine.

        :param any|None input_values: any input values for the question
        :param octue.resources.manifest.Manifest|None input_manifest: an input manifest of any datasets needed for the question
        :param list(dict)|None children: a list of children for the child to use instead of its default children (if it uses children). These should be in the same format as in an app's app configuration file and have the same keys.
        :param bool subscribe_to_logs: if `True`, subscribe to logs from the child and handle them with the local log handlers
        :param bool allow_local_files: if `True`, allow the input manifest to contain references to local files - this should only be set to `True` if the child will have access to these local files
        :param callable|None handle_monitor_message: a function to handle monitor messages (e.g. send them to an endpoint for plotting or displaying) - this function should take a single JSON-compatible python primitive as an argument (note that this could be an array or object)
        :param bool record_messages: if `True`, record messages received from the child in the `received_messages` property
        :param bool allow_save_diagnostics_data_on_crash: if `True`, allow the input values and manifest (and its datasets) to be saved by the child if it fails while processing them
        :param str|None question_uuid: the UUID to use for the question if a specific one is needed; a UUID is generated if not
        :param float timeout: time in seconds to wait for an answer before raising a timeout error
        :param float|int delivery_acknowledgement_timeout: how long in seconds to wait for a delivery acknowledgement before aborting
        :param float|int maximum_heartbeat_interval: the maximum amount of time (in seconds) allowed between child heartbeats before an error is raised
        :raise TimeoutError: if the timeout is exceeded while waiting for an answer
        :return dict: a dictionary containing the keys "output_values" and "output_manifest"
        """
        subscription, _ = self._service.ask(
            service_id=self.id,
            input_values=input_values,
            input_manifest=input_manifest,
            children=children,
            subscribe_to_logs=subscribe_to_logs,
            allow_local_files=allow_local_files,
            allow_save_diagnostics_data_on_crash=allow_save_diagnostics_data_on_crash,
            question_uuid=question_uuid,
            timeout=timeout,
        )

        return self._service.wait_for_answer(
            subscription=subscription,
            handle_monitor_message=handle_monitor_message,
            record_messages=record_messages,
            service_name=self.id,
            timeout=timeout,
            maximum_heartbeat_interval=maximum_heartbeat_interval,
        )

    def ask_multiple(self, *questions):
        """Ask the child multiple questions in parallel and wait for the answers. Each question should be provided as a
        dictionary of `Child.ask` keyword arguments. An error is raised and no answers are returned if any of the
        individual questions raise an error.

        :param questions: any number of questions provided as dictionaries of arguments to the `Child.ask` method
        :raises Exception: if any question raises an error.
        :return list: the answers to the questions in the same order as the questions
        """

        def ask(question):
            return self.ask(**question)

        max_workers = min(32, len(questions))

        with concurrent.futures.ThreadPoolExecutor(max_workers=max_workers) as executor:
            return list(executor.map(ask, questions))



import json
import logging

import coolname

import twined.exceptions
from octue.cloud import storage
from octue.exceptions import InvalidMonitorMessage
from octue.mixins import Hashable, Identifiable, Labelable, Serialisable, Taggable
from octue.resources.manifest import Manifest
from octue.utils.encoders import OctueJSONEncoder
from octue.utils.threads import RepeatingTimer
from twined import ALL_STRANDS, Twine


logger = logging.getLogger(__name__)


HASH_FUNCTIONS = {
    "configuration_values": Hashable.hash_non_class_object,
    "configuration_manifest": lambda manifest: manifest.hash_value,
    "input_values": Hashable.hash_non_class_object,
    "input_manifest": lambda manifest: manifest.hash_value,
}

# Map strand names to class which we expect Twined to instantiate for us
CLASS_MAP = {"configuration_manifest": Manifest, "input_manifest": Manifest, "output_manifest": Manifest}


class Analysis(Identifiable, Serialisable, Labelable, Taggable):
    """A class representing a scientific or computational analysis. It holds references to all configuration, input, and
    output data, logs, connections to child services, credentials, etc. It's essentially the "Internal API" for your
    service - a single point of contact where you can get or update anything you need.

    An ``Analysis`` instance is automatically provided to the app in an Octue service when a question is received. Its
    attributes include every strand that can be added to a ``Twine``, although only the strands specified in the
    service's twine will be non-``None``. Incoming data is validated before it's added to the analysis.

    All input and configuration attributes are hashed using a `BLAKE3 hash <https://github.com/BLAKE3-team/BLAKE3>`_ so
    the inputs and configuration that produced a given output in your app can always be verified. These hashes exist on
    the following attributes:

    -   ``input_values_hash``
    -   ``input_manifest_hash``
    -   ``configuration_values_hash``
    -   ``configuration_manifest_hash``

    If a strand is ``None``, so will its corresponding hash attribute be. The hash of a datafile is the hash of its
    file, while the hash of a manifest or dataset is the cumulative hash of the files it refers to.

    :param twined.Twine|dict|str twine: the twine, dictionary defining a twine, or path to "twine.json" file defining the service's data interface
    :param callable|None handle_monitor_message: an optional function for sending monitor messages to the parent that requested the analysis
    :param any configuration_values: the configuration values for the analysis - this can be expressed as a python primitive (e.g. dict), a path to a JSON file, or a JSON string.
    :param octue.resources.manifest.Manifest configuration_manifest: a manifest of configuration datasets for the analysis if required
    :param any input_values: the input values for the analysis - this can be expressed as a python primitive (e.g. dict), a path to a JSON file, or a JSON string.
    :param octue.resources.manifest.Manifest input_manifest: a manifest of input datasets for the analysis if required
    :param any output_values: any output values the analysis produces
    :param octue.resources.manifest.Manifest output_manifest: a manifest of output dataset from the analysis if it produces any
    :param dict children: a mapping of string key to ``Child`` instance for all the children used by the service
    :param str id: Optional UUID for the analysis
    :return None:
    """

    def __init__(self, twine, handle_monitor_message=None, **kwargs):
        if isinstance(twine, Twine):
            self.twine = twine
        else:
            self.twine = Twine(source=twine)

        self._handle_monitor_message = handle_monitor_message

        strand_kwargs = {name: kwargs.pop(name, None) for name in ALL_STRANDS}

        # Values strands.
        self.configuration_values = strand_kwargs.get("configuration_values", None)
        self.input_values = strand_kwargs.get("input_values", None)
        self.output_values = strand_kwargs.get("output_values", None)

        # Manifest strands.
        self.configuration_manifest = strand_kwargs.get("configuration_manifest", None)
        self.input_manifest = strand_kwargs.get("input_manifest", None)
        self.output_manifest = strand_kwargs.get("output_manifest", None)

        # Other strands.
        self.children = strand_kwargs.get("children", None)

        # Non-strands.
        self.output_location = kwargs.pop("output_location", None)

        self._calculate_strand_hashes(strands=strand_kwargs)
        self._periodic_monitor_message_sender_threads = []
        self._finalised = False
        super().__init__(**kwargs)

    @property
    def finalised(self):
        """Check whether the analysis has been finalised (i.e. whether its outputs have been validated and, if an output
        manifest is produced, its datasets uploaded).

        :return bool:
        """
        return self._finalised

    def send_monitor_message(self, data):
        """Send a monitor message to the parent that requested the analysis.

        :param any data: any JSON-compatible data structure
        :return None:
        """
        try:
            self.twine.validate_monitor_message(source=data)
        except twined.exceptions.InvalidValuesContents as e:
            raise InvalidMonitorMessage(e)

        if self._handle_monitor_message is None:
            logger.warning("Attempted to send a monitor message but no handler is specified.")
            return

        self._handle_monitor_message(data)

    def set_up_periodic_monitor_message(self, create_monitor_message, period=60):
        """Set up a periodic monitor message that sends up-to-date data once per period.

        :param callable create_monitor_message: a callable that takes no arguments and returns a new up-to-date monitor message to send each time it's called
        :param int|float period: the repetition period in seconds
        :return None:
        """
        thread = RepeatingTimer(
            interval=period,
            function=lambda: self.send_monitor_message(data=create_monitor_message()),
        )

        thread.daemon = True
        thread.start()

        self._periodic_monitor_message_sender_threads.append(thread)
        logger.info("Periodic monitor message set up to send every %ss.", period)

    def finalise(self, upload_output_datasets_to=None):
        """Validate the output values and output manifest and, if the analysis produced an output manifest, upload its
        output datasets to a unique subdirectory within the analysis's output location. This output location can be
        overridden by providing a different cloud path via the `upload_output_datasets_to` parameter. Either way, the
        dataset paths in the output manifest are replaced with signed URLs for easier, expiring access.

        :param str|None upload_output_datasets_to: If not provided but an output location was provided at instantiation, upload any output datasets into a unique subdirectory within this output location; if provided, upload into this location instead. The output manifest is updated with the upload locations.
        :return None:
        """
        serialised_strands = {"output_values": None, "output_manifest": None}

        if self.output_values:
            serialised_strands["output_values"] = json.dumps(self.output_values, cls=OctueJSONEncoder)

        if self.output_manifest:
            serialised_strands["output_manifest"] = self.output_manifest.serialise()

        self.twine.validate(**serialised_strands)
        self._finalised = True
        logger.info("Validated output values and output manifest against the twine.")

        # Use a unique subdirectory in the output location given at instantiation (if given) if no
        # `upload_output_datasets_to` is provided.
        if self.output_location and not upload_output_datasets_to:
            upload_output_datasets_to = storage.path.join(self.output_location, coolname.generate_slug())

        # If there isn't both an output manifest and upload location, nothing is uploaded.
        if not (upload_output_datasets_to and hasattr(self, "output_manifest")):
            return

        for name, dataset in self.output_manifest.datasets.items():
            dataset.upload(cloud_path=storage.path.join(upload_output_datasets_to, name))

        self.output_manifest.use_signed_urls_for_datasets()

        logger.info("Uploaded output datasets to %r.", upload_output_datasets_to)

    def _calculate_strand_hashes(self, strands):
        """Calculate the hashes of the strands specified in the HASH_FUNCTIONS constant.

        :param dict strands: strand names mapped to strand data
        :return None:
        """
        for strand_name, strand_data in strands.items():
            if strand_name in HASH_FUNCTIONS:
                strand_hash_name = f"{strand_name}_hash"

                if strand_data is not None:
                    setattr(self, strand_hash_name, HASH_FUNCTIONS[strand_name](strand_data))
                else:
                    setattr(self, strand_hash_name, None)



import logging
from urllib.parse import urlparse

from octue.cloud import storage
from octue.exceptions import CloudLocationNotSpecified


logger = logging.getLogger(__name__)


class CloudPathable:
    """A mixin providing cloud-related properties and functions common to `octue` resources."""

    _CLOUD_PATH_ATTRIBUTE_NAME = "path"

    @property
    def exists_in_cloud(self):
        """Return `True` if the instance exists in the cloud.

        :return bool:
        """
        if self.__cloud_path:
            return storage.path.is_cloud_path(self.__cloud_path)
        return None

    @property
    def cloud_protocol(self):
        """Get the cloud protocol of the instance if it exists in the cloud (e.g. "gs" for a cloud path of
        "gs://my-bucket/my-file.txt").

        :return str|None:
        """
        if not self.exists_in_cloud:
            return None
        return urlparse(self.__cloud_path).scheme

    @property
    def bucket_name(self):
        """Get the name of the bucket the instance exists in if it exists in the cloud.

        :return str|None:
        """
        if self.exists_in_cloud:
            return storage.path.split_bucket_name_from_cloud_path(self.__cloud_path)[0]
        return None

    @property
    def path_in_bucket(self):
        """Get the path of the instance in its bucket if it exists in the cloud.

        :return str|None:
        """
        if self.exists_in_cloud:
            return storage.path.split_bucket_name_from_cloud_path(self.__cloud_path)[1]
        return None

    def _set_cloud_location(self, cloud_path=None):
        """Set the cloud location for the instance. If no cloud path is given, the current cloud path is used if one has
        previously been set. Otherwise, an error is raised.

        :param str|None cloud_path:
        :raise octue.exceptions.CloudLocationNotSpecified: if an exact cloud location isn't provided and isn't available implicitly (i.e. the instance wasn't loaded from the cloud previously)
        :return None:
        """
        cloud_path = cloud_path or self.__cloud_path

        if not cloud_path:
            self._raise_cloud_location_error()

        if cloud_path != self.__cloud_path:
            logger.debug("Updating cloud location for %r from %r to %r.", self, self.__cloud_path, cloud_path)

        setattr(self, self._CLOUD_PATH_ATTRIBUTE_NAME, cloud_path)

    def _raise_cloud_location_error(self):
        """Raise an error indicating that the cloud location of the instance has not yet been specified.

        :raise CloudLocationNotSpecified:
        :return None:
        """
        raise CloudLocationNotSpecified(
            f"{self!r} wasn't previously loaded from the cloud so doesn't have an implicit cloud location - please "
            f"specify its cloud path."
        )

    @property
    def __cloud_path(self):
        """Get the cloud path for the instance according to the class variable `_CLOUD_PATH_ATTRIBUTE_NAME`.

        :return str:
        """
        return getattr(self, self._CLOUD_PATH_ATTRIBUTE_NAME)



import base64
import concurrent.futures
import datetime
import functools
import json
import logging
import threading
import uuid

import google.api_core.exceptions
import pkg_resources
from google.api_core import retry
from google.cloud import pubsub_v1

import octue.exceptions
from octue.cloud.pub_sub import Subscription, Topic
from octue.cloud.pub_sub.logging import GooglePubSubHandler
from octue.cloud.pub_sub.message_handler import OrderedMessageHandler
from octue.cloud.service_id import convert_service_id_to_pub_sub_form, create_service_sruid, validate_service_sruid
from octue.compatibility import warn_if_incompatible
from octue.utils.decoders import OctueJSONDecoder
from octue.utils.encoders import OctueJSONEncoder
from octue.utils.exceptions import convert_exception_to_primitives
from octue.utils.objects import get_nested_attribute
from octue.utils.threads import RepeatingTimer


logger = logging.getLogger(__name__)
send_message_lock = threading.Lock()

DEFAULT_NAMESPACE = "default"
ANSWERS_NAMESPACE = "answers"

# Switch message batching off by setting `max_messages` to 1. This minimises latency and is recommended for
# microservices publishing single messages in a request-response sequence.
BATCH_SETTINGS = pubsub_v1.types.BatchSettings(max_bytes=10 * 1000 * 1000, max_latency=0.01, max_messages=1)


class Service:
    """An Octue service that can be used as a data service or digital twin in one of two modes:

    - As a child accepting questions (input values and manifests) from parents, running them through its app, and
      responding with an answer
    - As a parent asking questions to children in the above mode

    Services communicate entirely via Google Pub/Sub and can ask and/or respond to questions from any other service that
    has a corresponding topic on Google Pub/Sub.

    :param octue.resources.service_backends.ServiceBackend backend: the object representing the type of backend the service uses
    :param str|None service_id: a unique ID to give to the service (any string); a UUID is generated if none is given
    :param callable|None run_function: the function the service should run when it is called
    :param str|None name: an optional name to use for the service to override its ID in its string representation
    :return None:
    """

    def __init__(self, backend, service_id=None, run_function=None, name=None):
        if service_id is None:
            self.id = create_service_sruid(namespace=DEFAULT_NAMESPACE, name=str(uuid.uuid4()))

        # Raise an error if the service ID is some kind of falsey object that isn't `None`.
        elif not service_id:
            raise ValueError(f"`service_id` should be `None` or a non-falsey value; received {service_id!r} instead.")

        else:
            validate_service_sruid(service_id)
            self.id = service_id

        self.backend = backend
        self.run_function = run_function
        self.name = name
        self._pub_sub_id = convert_service_id_to_pub_sub_form(self.id)
        self._local_sdk_version = pkg_resources.get_distribution("octue").version
        self._publisher = None
        self._message_handler = None

    def __repr__(self):
        return f"<{type(self).__name__}({self.name or self.id!r})>"

    @property
    def publisher(self):
        """Get or instantiate the publisher client for the service. No publisher is instantiated until this property is
        called for the first time. This allows checking for the `GOOGLE_APPLICATION_CREDENTIALS` environment variable to
        be put off until it's needed.

        :return google.cloud.pubsub_v1.PublisherClient:
        """
        if not self._publisher:
            self._publisher = pubsub_v1.PublisherClient(batch_settings=BATCH_SETTINGS)

        return self._publisher

    @property
    def received_messages(self):
        """Get the messages received by the service from a child service while running the `wait_for_answer` method. If
        the `wait_for_answer` method hasn't been run, `None` is returned. If an empty list is returned, no messages have
        been received.

        :return list(dict)|None:
        """
        if self._message_handler:
            return self._message_handler.handled_messages
        return None

    def serve(self, timeout=None, delete_topic_and_subscription_on_exit=False, allow_existing=False, detach=False):
        """Start the service as a child, waiting to accept questions from any other Octue service using Google Pub/Sub
        on the same Google Cloud project. Questions are accepted, processed, and answered asynchronously.

        :param float|None timeout: time in seconds after which to shut down the child
        :param bool delete_topic_and_subscription_on_exit: if `True`, delete the child's topic and subscription on exiting serving mode
        :param bool allow_existing: if `True`, allow starting a service for which the topic and/or subscription already exists (indicating an existing service) - this connects this service to the existing service's topic and subscription
        :param bool detach: if `True`, detach from the subscription future. The future and subscriber are returned so can still be cancelled and closed manually. Note that the topic and subscription are not automatically deleted on exit if this option is chosen.
        :return (google.cloud.pubsub_v1.subscriber.futures.StreamingPullFuture, google.cloud.pubsub_v1.SubscriberClient):
        """
        logger.info("Starting %r.", self)

        topic = Topic(name=self._pub_sub_id, project_name=self.backend.project_name)

        subscription = Subscription(
            name=self._pub_sub_id,
            topic=topic,
            project_name=self.backend.project_name,
            expiration_time=None,
        )

        try:
            topic.create(allow_existing=allow_existing)
            subscription.create(allow_existing=allow_existing)
        except google.api_core.exceptions.AlreadyExists:
            raise octue.exceptions.ServiceAlreadyExists(f"A service with the ID {self.id!r} already exists.")

        subscriber = pubsub_v1.SubscriberClient()

        try:
            future = subscriber.subscribe(subscription=subscription.path, callback=self.answer)

            logger.info(
                "You can now ask this service questions at %r using the `octue.resources.Child` class.",
                self.id,
            )

            # If not detaching, keep answering questions until the subscriber times out (or forever if there's no
            # timeout).
            if not detach:
                try:
                    future.result(timeout=timeout)
                except (TimeoutError, concurrent.futures.TimeoutError, KeyboardInterrupt):
                    future.cancel()

        finally:
            # If not detaching, delete the topic and subscription deletion if required and close the subscriber.
            if not detach:
                if delete_topic_and_subscription_on_exit:
                    try:
                        if subscription.creation_triggered_locally:
                            subscription.delete()

                        if topic.creation_triggered_locally:
                            topic.delete()

                    except Exception:
                        logger.error("Deletion of topic and/or subscription %r failed.", topic.name)

                subscriber.close()

        return future, subscriber

    def answer(self, question, answer_topic=None, heartbeat_interval=120, timeout=30):
        """Answer a question from a parent - i.e. run the child's app on the given data and return the output values.
        Answers conform to the output values and output manifest schemas specified in the child's Twine file.

        :param dict|Message question:
        :param octue.cloud.pub_sub.topic.Topic|None answer_topic: provide if messages need to be sent to the parent from outside the `Service` instance (e.g. in octue.cloud.deployment.google.cloud_run.flask_app)
        :param int|float heartbeat_interval: the time interval, in seconds, at which to send heartbeats
        :param float|None timeout: time in seconds to keep retrying sending of the answer once it has been calculated
        :raise Exception: if any exception arises during running analysis and sending its results
        :return None:
        """
        (
            data,
            question_uuid,
            forward_logs,
            parent_sdk_version,
            allow_save_diagnostics_data_on_crash,
        ) = self._parse_question(question)

        topic = answer_topic or self.instantiate_answer_topic(question_uuid)
        self._send_delivery_acknowledgment(topic)

        heartbeater = RepeatingTimer(
            interval=heartbeat_interval,
            function=self._send_heartbeat,
            kwargs={"topic": topic},
        )

        heartbeater.daemon = True
        heartbeater.start()

        try:
            if forward_logs:
                analysis_log_handler = GooglePubSubHandler(
                    message_sender=self._send_message,
                    topic=topic,
                    analysis_id=question_uuid,
                )
            else:
                analysis_log_handler = None

            analysis = self.run_function(
                analysis_id=question_uuid,
                input_values=data["input_values"],
                input_manifest=data["input_manifest"],
                children=data.get("children"),
                analysis_log_handler=analysis_log_handler,
                handle_monitor_message=functools.partial(self._send_monitor_message, topic=topic),
                allow_save_diagnostics_data_on_crash=allow_save_diagnostics_data_on_crash,
            )

            if analysis.output_manifest is None:
                serialised_output_manifest = None
            else:
                serialised_output_manifest = analysis.output_manifest.serialise()

            self._send_message(
                {
                    "type": "result",
                    "output_values": analysis.output_values,
                    "output_manifest": serialised_output_manifest,
                },
                topic=topic,
                timeout=timeout,
            )

            heartbeater.cancel()
            logger.info("%r answered question %r.", self, question_uuid)

        except BaseException as error:  # noqa
            heartbeater.cancel()
            warn_if_incompatible(child_sdk_version=self._local_sdk_version, parent_sdk_version=parent_sdk_version)
            self.send_exception(topic, timeout)
            raise error

    def ask(
        self,
        service_id,
        input_values=None,
        input_manifest=None,
        children=None,
        subscribe_to_logs=True,
        allow_local_files=False,
        allow_save_diagnostics_data_on_crash=True,
        question_uuid=None,
        push_endpoint=None,
        timeout=86400,
    ):
        """Ask a child a question (i.e. send it input values for it to analyse and produce output values for) and return
        a subscription to receive its answer on. The input values and manifest must conform to the schemas in the
        child's Twine file.

        :param str service_id: the ID of the child to ask the question to
        :param any|None input_values: any input values for the question
        :param octue.resources.manifest.Manifest|None input_manifest: an input manifest of any datasets needed for the question
        :param list(dict)|None children: a list of children for the child to use instead of its default children (if it uses children). These should be in the same format as in an app's app configuration file and have the same keys.
        :param bool subscribe_to_logs: if `True`, subscribe to the child's logs and handle them with the local log handlers
        :param bool allow_local_files: if `True`, allow the input manifest to contain references to local files - this should only be set to `True` if the child will be able to access these local files
        :param bool allow_save_diagnostics_data_on_crash: if `True`, allow the input values and manifest (and its datasets) to be saved by the child if it fails while processing them
        :param str|None question_uuid: the UUID to use for the question if a specific one is needed; a UUID is generated if not
        :param str|None push_endpoint: if answers to the question should be pushed to an endpoint, provide its URL here; if they should be pulled, leave this as `None`
        :param float|None timeout: time in seconds to keep retrying sending the question
        :return (octue.cloud.pub_sub.subscription.Subscription, str): the answer subscription and question UUID
        """
        validate_service_sruid(service_id)

        if not allow_local_files:
            if (input_manifest is not None) and (not input_manifest.all_datasets_are_in_cloud):
                raise octue.exceptions.FileLocationError(
                    "All datasets of the input manifest and all files of the datasets must be uploaded to the cloud "
                    "before asking a service to perform an analysis upon them. The manifest must then be updated with "
                    "the new cloud locations."
                )

        pub_sub_service_id = convert_service_id_to_pub_sub_form(service_id)
        question_topic = Topic(name=pub_sub_service_id, project_name=self.backend.project_name)

        if not question_topic.exists(timeout=timeout):
            raise octue.exceptions.ServiceNotFound(f"Service with ID {service_id!r} cannot be found.")

        question_uuid = question_uuid or str(uuid.uuid4())

        answer_topic = self.instantiate_answer_topic(question_uuid, pub_sub_service_id)
        answer_topic.create(allow_existing=False)

        answer_subscription = Subscription(
            name=answer_topic.name,
            topic=answer_topic,
            project_name=self.backend.project_name,
            push_endpoint=push_endpoint,
        )
        answer_subscription.create(allow_existing=False)

        if input_manifest is not None:
            input_manifest.use_signed_urls_for_datasets()
            input_manifest = input_manifest.to_primitive()

        self._send_message(
            {"input_values": input_values, "input_manifest": input_manifest, "children": children},
            topic=question_topic,
            question_uuid=question_uuid,
            forward_logs=subscribe_to_logs,
            allow_save_diagnostics_data_on_crash=allow_save_diagnostics_data_on_crash,
        )

        logger.info("%r asked a question %r to service %r.", self, question_uuid, service_id)
        return answer_subscription, question_uuid

    def wait_for_answer(
        self,
        subscription,
        handle_monitor_message=None,
        record_messages=True,
        service_name="REMOTE",
        timeout=60,
        maximum_heartbeat_interval=300,
    ):
        """Wait for an answer to a question on the given subscription, deleting the subscription and its topic once
        the answer is received.

        :param octue.cloud.pub_sub.subscription.Subscription subscription: the subscription for the question's answer
        :param callable|None handle_monitor_message: a function to handle monitor messages (e.g. send them to an endpoint for plotting or displaying) - this function should take a single JSON-compatible python primitive as an argument (note that this could be an array or object)
        :param bool record_messages: if `True`, record messages received from the child in the `received_messages` attribute
        :param str service_name: a name by which to refer to the child subscribed to (used for labelling its log messages if subscribed to)
        :param float|None timeout: how long in seconds to wait for an answer before raising a `TimeoutError`
        :param float|int delivery_acknowledgement_timeout: how long in seconds to wait for a delivery acknowledgement before aborting
        :param float|int maximum_heartbeat_interval: the maximum amount of time (in seconds) allowed between child heartbeats before an error is raised
        :raise TimeoutError: if the timeout is exceeded
        :return dict: dictionary containing the keys "output_values" and "output_manifest"
        """
        if subscription.is_push_subscription:
            raise octue.exceptions.PushSubscriptionCannotBePulled(
                f"{subscription.path!r} is a push subscription so it cannot be waited on for an answer. Please check "
                f"its push endpoint at {subscription.push_endpoint!r}."
            )

        self._message_handler = OrderedMessageHandler(
            subscription=subscription,
            receiving_service=self,
            handle_monitor_message=handle_monitor_message,
            service_name=service_name,
            record_messages=record_messages,
        )

        try:
            return self._message_handler.handle_messages(
                timeout=timeout,
                maximum_heartbeat_interval=maximum_heartbeat_interval,
            )

        finally:
            subscription.delete()

    def instantiate_answer_topic(self, question_uuid, service_id=None):
        """Instantiate the answer topic for the given question UUID and child service ID.

        :param str question_uuid:
        :param str|None service_id: the ID of the child to ask the question to in Pub/Sub form
        :return octue.cloud.pub_sub.topic.Topic:
        """
        return Topic(
            name=".".join((service_id or self._pub_sub_id, ANSWERS_NAMESPACE, question_uuid)),
            project_name=self.backend.project_name,
        )

    def send_exception(self, topic, timeout=30):
        """Serialise and send the exception being handled to the parent.

        :param octue.cloud.pub_sub.topic.Topic topic:
        :param float|None timeout: time in seconds to keep retrying sending of the exception
        :return None:
        """
        exception = convert_exception_to_primitives()
        exception_message = f"Error in {self!r}: {exception['message']}"

        self._send_message(
            {
                "type": "exception",
                "exception_type": exception["type"],
                "exception_message": exception_message,
                "traceback": exception["traceback"],
            },
            topic=topic,
            timeout=timeout,
        )

    def _send_message(self, message, topic, timeout=30, **attributes):
        """Send a JSON-serialised message to the given topic with optional message attributes.

        :param dict message: JSON-serialisable data to send as a message
        :param octue.cloud.pub_sub.topic.Topic topic: the Pub/Sub topic to send the message to
        :param int|float timeout: the timeout for sending the message in seconds
        :param attributes: key-value pairs to attach to the message - the values must be strings or bytes
        :return None:
        """
        with send_message_lock:
            attributes["octue_sdk_version"] = self._local_sdk_version

            # This would be better placed in the Pub/Sub message's attributes but has been left in `message` for
            # inter-service backwards compatibility.
            message["message_number"] = topic.messages_published
            converted_attributes = {}

            for key, value in attributes.items():
                if isinstance(value, bool):
                    value = str(int(value))
                elif isinstance(value, (int, float)):
                    value = str(value)

                converted_attributes[key] = value

            self.publisher.publish(
                topic=topic.path,
                data=json.dumps(message, cls=OctueJSONEncoder).encode(),
                retry=retry.Retry(deadline=timeout),
                **converted_attributes,
            )

            topic.messages_published += 1

    def _send_delivery_acknowledgment(self, topic, timeout=30):
        """Send an acknowledgement of question receipt to the parent.

        :param octue.cloud.pub_sub.topic.Topic topic: topic to send the acknowledgement to
        :param float timeout: time in seconds after which to give up sending
        :return None:
        """
        self._send_message(
            {
                "type": "delivery_acknowledgement",
                "delivery_time": str(datetime.datetime.now()),
            },
            topic=topic,
            timeout=timeout,
        )

        logger.info("%r acknowledged receipt of question.", self)

    def _send_heartbeat(self, topic, timeout=30):
        """Send a heartbeat to the parent, indicating that the service is alive.

        :param octue.cloud.pub_sub.topic.Topic topic: topic to send the heartbeat to
        :param float timeout: time in seconds after which to give up sending
        :return None:
        """
        self._send_message(
            {
                "type": "heartbeat",
                "time": str(datetime.datetime.now()),
            },
            topic=topic,
            timeout=timeout,
        )

        logger.debug("Heartbeat sent by %r.", self)

    def _send_monitor_message(self, data, topic, timeout=30):
        """Send a monitor message to the parent.

        :param any data: the data to send as a monitor message
        :param octue.cloud.pub_sub.topic.Topic topic: the topic to send the message to
        :param float timeout: time in seconds to retry sending the message
        :return None:
        """
        self._send_message(
            {
                "type": "monitor_message",
                "data": json.dumps(data, cls=OctueJSONEncoder),
            },
            topic=topic,
            timeout=timeout,
        )

        logger.debug("Monitor message sent by %r.", self)

    def _parse_question(self, question):
        """Parse a question in the Google Cloud Pub/Sub or Google Cloud Run format.

        :param dict|Message question:
        :return (dict, str, bool, str|None, bool):
        """
        logger.info("%r received a question.", self)

        try:
            # Parse question directly from Pub/Sub or Dataflow.
            data = json.loads(question.data.decode(), cls=OctueJSONDecoder)

            # Acknowledge it if it's directly from Pub/Sub
            if hasattr(question, "ack"):
                question.ack()

        except Exception:
            # Parse question from Google Cloud Run.
            data = json.loads(base64.b64decode(question["data"]).decode("utf-8").strip(), cls=OctueJSONDecoder)

        # Keep backwards compatibility with questions from Octue services running `octue<0.41.1`.
        if isinstance(data["input_manifest"], str):
            data["input_manifest"] = json.loads(data["input_manifest"], cls=OctueJSONDecoder)

        question_uuid = get_nested_attribute(question, "attributes.question_uuid")
        forward_logs = bool(int(get_nested_attribute(question, "attributes.forward_logs")))

        try:
            parent_sdk_version = get_nested_attribute(question, "attributes.octue_sdk_version")
        except AttributeError:
            parent_sdk_version = None

        try:
            allow_save_diagnostics_data_on_crash = get_nested_attribute(
                question,
                "attributes.allow_save_diagnostics_data_on_crash",
            )
        except AttributeError:
            allow_save_diagnostics_data_on_crash = False

        logger.info("%r parsed the question successfully.", self)
        return data, question_uuid, forward_logs, parent_sdk_version, allow_save_diagnostics_data_on_crash
